import numpy as np
from cython.parallel import prange

def update_lloyd(
        sq_distances, 
        double[:, ::1] kernel_matrix,
        long[::1] labels, 
        long n_clusters,
        long[::1] cluster_sizes):
    '''
    Computes squared dists from points to centers in feature space
 
    First, checks for empty cluster and fills them with a random 
    element. Then computes the sq_dist from each point to each center.
    
    Parameters
    ----------
    sq_distances: ndarray of shape(n_samples, n_clusters), dtype=np.double
        sq_distances[x, :] = K(x,x). Could be zeros but then inertia
        quality measure would not be exact.
    kernel_matrix: ndarray of shape(n_samples, n_samples), dtype=double
    labels: ndarray of shape(n_samples), dtype=long
    n_clusters: long

    Returns
    -------
    sq_distances: ndarray of shape(n_samples, n_clusters)
    inner_sums: ndarray of shape(n_clusters)
        sum of pairwise sq_dists in each cluster, used
        for prediction
    cluster_sizes: ndarray of shape(n_clusters)
    
    Notes
    -----
    Numpy ndarray can be passed for memoryview.
    sq_distances parameter deliberately not typed because
    of np broadcast addition in loop.
    TODO measure speed if typed explicitely.
    '''         
    cdef Py_ssize_t cluster # loop index

    sq_distances = np.asarray(sq_distances, dtype=np.double)

    # labels, cluster_sizes = fill_empty_clusters(labels, n_clusters, return_sizes=True)
    outer_sums, inner_sums = _calc_sums_full(kernel_matrix, labels, n_clusters)
    
    for cluster in range(n_clusters):
        size = cluster_sizes[cluster]
        outer_sum = outer_sums[:, cluster]
        inner_sum = inner_sums[cluster] 
        sq_distances[:,  cluster] += (-2 * outer_sum / size 
                                     + inner_sum / size**2)
    return np.asarray(sq_distances), np.asarray(inner_sums), np.asarray(cluster_sizes)


def _calc_sums_full(double[:, ::1] kernel_matrix, long[::1] labels, long n_clusters):
    '''
    Helper function for distance calculation.

    Parameters
    ----------
    kernel_matrix: ndarray of shape(n_samples, n_samples), dtype=double
        pairwise kernel matrix of input data
    labels: ndarray of shape(n_samples), dtype=long
    n_clusters: int

    Returns
    -------
    outer_sums: ndarray of shape(n_samples, n_clusters)
        outer_sums[i, j] == sum(K(data[i], data[x])) for x in cluster j
        In words, the sum of the kernel of datapoint X[i] and all datapoints
        in cluster number j.
    inner_sums: ndarray of shape(n_clusters)
        "inner 'distance' sum"; 
        inner_sums[i] = sum(K(x,y)) for all x, y in cluster i
    '''


    cdef:
        Py_ssize_t size = kernel_matrix.shape[0]
        double[:, ::1] inner_sum = np.zeros((size, n_clusters), dtype=np.double)
        double[:, ::1] outer_sum = np.zeros((size, n_clusters), dtype=np.double)
        Py_ssize_t i, j
        int label_j

    assert size == kernel_matrix.shape[1]

    for i in prange(size, nogil=True):
        for j in range(size):
            label_j = labels[j]
            outer_sum[i, label_j] += kernel_matrix[i, j]      
            if labels[i] == label_j:
                inner_sum[i, label_j] += kernel_matrix[i, j]
    return np.asarray(outer_sum), np.sum(inner_sum, axis=0)


import numpy as np
from KKMeans.lloyd import update_lloyd
from KKMeans.utils import calc_sq_distances, fill_empty_clusters
from KKMeans.kernels import build_kernel_matrix
from KKMeans.quality import avg_silhouette
from KKMeans.elkan import update_elkan, start_elkan
from matplotlib import pyplot as plt

class KKMeans():
    '''
    Kernel K-Means Clustering.

    Read more in the thesis. TODO
    Designed to mimic sklearn.cluster.KMeans
    limitations: kernel matrix

    Parameters 
    ----------

    n_clusters: int, default=8
        The number of clusters to form.

    init: {kmeans++, random, truerandom} or arraylike of shape(n_clusters, n_features), default = kmeans++
        Initialization method. 
        kmeans++: heuristic to choose "good" starting centers
            For more, see: thesis or
            Arthur and Vassilvitskii, Kmeans++ the 
            advantage of carefull seeding, 2007
        random: Randomly takes existing datapoints as centers
        truerandom: assigns random labels to each datapoint
    
        If arraylike is passed, init is cast to ndarray of
        doubles and its content is used as centers.

    n_init: int, default=3 
        The number of clusterings computed. As KKMeans is 
        a heuristic that converges to local optima, 
        mutliple runs are beneficial. 
    
    max_iter: int, default=300
        Maximal amount of iterations for each run.
    
    tol: float, default = 0.0001    
        Max difference of quality-metric between iterations
        to declare run for converged.
        If 0., run converges only when labels do not change
        or max_iter is reached.
    
    q_metrics: {inertia, silhouette}, default=inertia
        Quality metric by which each clustering is measured.

        inertia:  also called variance, sum of squared distances 
            from each point to its assigned center. Objective function in 
            (Kernel) K-means. 

        silhouette: Measures how "ambiguous" the assignment of a point
            to its cluster is; if the loss function would not change
            if it were assigned to the next best cluster, the silhouette 
            of said point would be 0.
            -1 <= silhouette <= 1
            For elkan, this does not work as elkan does not calculate
            the distance to every cluster centers.
            calculated via:
            a(x) = d(x, centers[x])
            b(x) = min_y(d(x, centers[y])) where y != x
            silhouette(x) = (b(x) - a(x))/max{a(x), b(x)}
            This is then averaged for every point.  
    
    verbose: bool, default=False
    
    rng: int, default=None
        Seed which will be passed to np.random.defaul_rng
        to create random generator.
        When none, "fresh, unpredictable entropy will be pulled from the OS"
            - np random generator docpage.
    
    algorithm: {"lloyd", "elkan"}, default="lloyd"
        The algorithm which will be used for commputation.
        lloyd is the standard heuristic used to solve the k-means   
        problem. This is a simplified version of Elkan's algorithm
        that does not need to calculate pairwise cluster distances.
        The original elkan (wiht pw cluster distances) is slower
        than lloyd's when utilizing a kernel matrix.

    kernel: {"linear", "rbf", "polynomial", "sigmoid", "gaussian", "laplacian"}, default="linear"
        Which kernel shall be used to compute the kernel matrix.
        Gaussian is equal to rbf with gamma = variance**-2
        Linear calculates a regular kmeans clustering.

    **kwargs: 
        Arguments which are passed to the kernel function.
        Dependant on kernel used

    Attributes
    ----------

    labels_: ndarray of shape (n_samples)
        labels of each point
    
    inner_sums_: ndarray of shape (n_clusters)
        Mathematical term used for prediction, stored
        to avoid recomputing/storing full kernel matrix.
        =SUM(K(x, x')) -> for each x, x' in cluster C
            -> for each cluster C
    
    cluster_sizes_: ndarray of shape(n_clusters)
        amount of points in each clusters
    
    X_: ndarray of shape(n_samples, n_features)
        The data which was passed to be clustered.

    quality_: float
        Quality measured by the passed q_metric.
        If tol=0 only gets measured at the end of a "run".

    '''
    def __init__(self, n_clusters=8, init="kmeans++", n_init=3,
                 max_iter=300, tol=1e-4, q_metric="inertia", verbose=False,
                 rng=None, algorithm="lloyd", kernel="linear", **kwargs):
        '''
        Ctor for KKMeans, more details in class docstring. 
        Calls validation function after initializing the class parameters,
        which can raise exceptions.
        
        Parameters:
        ----------
        n_clusters: int, default=8
        init: {"kmeans++", "random", "truerandom"}, default = "kmeans++"
        n_init: int, default=3
        max_iter: int, default=300
        tol: float, default=0.0001
        q_metric: {inertia, silhouette}, default="inertia"
        verbose: bool, default=True
        rng: int, default=None
        algorithm: {"lloyd", "elkan"}, default="lloyd"
        kernel: {"linear", "rbf", "polynomial", "sigmoid", "gaussian", "laplacian"}, default="linear"

        '''
        self.n_clusters = n_clusters
        self.init = init
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.q_metric = q_metric
        self.verbose = bool(verbose)
        self.rng = np.random.default_rng(rng)
        self.algorithm = algorithm
        self.kernel = kernel
        self.kwargs = kwargs

        self._validate_params()

        # Class attributes that are used for information storage

        self.labels_ = None
        self.inner_sums_ = None
        self.cluster_sizes_ = None
        self.X_ = None
        self.quality_ = None
    

    def _validate_params(self):
        '''Checks for possible errors in the parameters, can raise exceptions'''
        self.__validate_n_clusters()
        self.__validate_init()
        self.__validate_n_init()
        self.__validate_max_iter()
        self.__validate_tol()
        self.__validate_q_metric()
        self.__validate_algorithm()
        self.__validate_kernel()
        if self.algorithm == "elkan" and self.q_metric == "silhouette":
            print("WARNING: using silhouette with elkan will most likely be inaccurate\
                  as elkan does not calculate exact distances to every center")

    def __validate_n_clusters(self):
        if not isinstance(self.n_clusters, int):
            raise ValueError("n_clusters must be int")
        if self.n_clusters < 1:
            raise ValueError("n_cluster must be >= 1")
    
    def __validate_init(self):
        if hasattr(self.init, "__iter__") and not isinstance(self.init, str):
            self.__validate_centers() 
        elif self.init not in ("kmeans++", "random", "truerandom") :
            raise NotImplementedError("Initialisation method not implemented")
    
    def __validate_centers(self):
        '''checks for invalid centers, can raise exceptions.'''
        self.init = np.asarray(self.init, dtype=np.double)
        if self.init.ndim != 2:
            raise ValueError("Given centers need to be 2-d array")
        if 0 in self.init.shape:
            raise ValueError("Given centers are empty")
        if self.init.shape[0] != self.n_clusters:
            raise ValueError("Amount of given centers must be equal to n_clusters")
        return self.init

    def __validate_n_init(self):
        if not isinstance(self.n_init, int):
            raise ValueError("n_init must be int")
        if self.n_init <= 0:
            raise ValueError("n_inits must be => 1")
    
    def __validate_max_iter(self):
        if not isinstance(self.max_iter, int):
            raise ValueError("max_iter must be int")
        if self.max_iter <= 0:
            raise ValueError("max_iter must be => 1")
        
    def __validate_algorithm(self):
        if self.algorithm not in ("lloyd", "elkan"):
            raise NotImplementedError(str(self.algorithm)+ " not implemented")
    
    def __validate_q_metric(self):
        if self.q_metric not in ("silhouette, inertia"):
            raise NotImplementedError("Quality metric not implemented")
    
    def __validate_tol(self):
        if not isinstance(self.tol, (int, float)):
            raise ValueError("tol must be numeric")
        if self.tol < 0:
            self.tol = 0
            print("Warning! tol < 0 passed, tol set to 0")

    def __validate_kernel(self):
        if self.kernel not in [
                "linear", "rbf", "polynomial", 
                "gaussian", "laplacian"]:
            raise ValueError("Invalid kernel provided.")
    

    
    def kernel_wrapper(self, X, Y=None):
        '''
        Wrapper function to compute kernel matrix
        
        If Y is None, Kernel is computed pairwise KM[i,j] = k(X[i], X[j])
        If Y is not None, KM[i,j] = k(X[i], Y[j])

        Parameters
        ----------
        X: arraylike of shape(*, n_features)
        Y: arraylike of shape(*, n_features) or None, default=None

        Returns
        -------
        Kernel Matrix, either of shape(len(X), len(X))
            or (len(X), len(Y))
        '''
        return build_kernel_matrix(X, Y, kernel=self.kernel, **self.kwargs)

    def fit(self, X):
        '''
        Fits the data X with the given parameters.

        Executes self.n_inits runs of self.algorithm,
        utlizing self.q_metric with self.tol to check for convergence.
        Safes the results in class attributes.

        Parameters
        ----------
        X: arraylike of shape(n_samples, n_features)
            The data which should be fit.
        '''
        X = np.ascontiguousarray(X, dtype=np.double)
        X = self._validate_data(X)
        kernel_matrix = self.kernel_wrapper(X)
        
        #Storage for each attribute to be safed per iteration
        labels_store = np.zeros((self.n_init, X.shape[0]), dtype=np.int_)
        quality_store = np.zeros(self.n_init, dtype=np.double)
        inner_sums_store = np.zeros((self.n_init, self.n_clusters), dtype=np.double)
        sizes_store = np.zeros((self.n_init, self.n_clusters), dtype=np.int_)
        
        for init in range(self.n_init):
            start_labels = self._init_labels(X, kernel_matrix)
            if self.algorithm == "lloyd":
                labels, quality, inner_sums, sizes = self.lloyd(kernel_matrix, start_labels)
            elif self.algorithm == "elkan":
                labels, quality, inner_sums, sizes = self.elkan(kernel_matrix, start_labels)
            else:
                raise ValueError(str(self.algorithm) + " algorithm not implemented; \
                                 this should have been caught in _check_params")
            labels_store[init] = labels
            quality_store[init] = quality
            inner_sums_store[init] = inner_sums
            sizes_store[init] = sizes

        best_init = self._get_best_init(quality_store)
        self.labels_ = labels_store[best_init]
        self.inner_sums_ = inner_sums_store[best_init]
        self.cluster_sizes_ = sizes_store[best_init]
        self.quality_ = quality_store[best_init]
        self.X_ = X
        
        self._out_finish(best_init)

    def _get_best_init(self, quality_store):
        """Returns best quality score given self.q_metric"""
        if self.q_metric == "inertia": # smalles is better
            return np.argmin(quality_store) 
        elif self.q_metric == "silhouette": # bigger is better
            return np.argmax(quality_store) 
        else:
            raise NotImplementedError(
                str(self.q_metric) 
                + " metric not implemented; \
                this should have been caught in _validate_params")

    def _out_finish(self, best_init):
        '''output at the end of self.fit'''
        print("Best " + self.q_metric +":", self.quality_,
              "Found at init:", best_init + 1)

    def _validate_data(self, X):
        """checks for invalidate data, can raise exceptions."""
        if len(X.shape) != 2:
            raise ValueError("X needs to be 2-d Array")
        if 0 in X.shape:
            raise ValueError("X is empty")
        if X.shape[0] <  self.n_clusters:
            raise ValueError("sample:cluster ratio needs to be at least one")
        return X
    
    def _init_labels(self, X, kernel_matrix):
        '''Assign labels to each datapoint by given method'''

        # hasattr(x, __iter__) to check if object may be arraylike 
        # if hasattr(self.init, "__iter__") and not isinstance(self.init, str): 
        if isinstance(self.init, np.ndarray) and self.init.dtype == np.double:
            return self._assign_to_centers(X, self.init)
        
        elif self.init == "random":
            centers = self.rng.choice(X, self.n_clusters)
            return self._assign_to_centers(X, centers)
        
        elif self.init == "truerandom":
            return self.rng.integers(0, self.n_clusters, len(X), dtype=np.int_)
        
        elif self.init == "kmeans++":
            return self.kmeanspp(X, kernel_matrix)
        
        raise NotImplementedError("Unknown initialisation method")
        
    def _assign_to_centers(self, X, centers):
        '''
        Assigns each datapoint to the closest given center.

        Distance is calculated via Kernel trick.
        In feature space, d(x, c) is calculated via:
        sqrt(K(x, x) - 2K(x, c) + K(c,c)), K being
        the kernel function.
        As 0 <= sqrt(x) < sqrt(y) => x < y the sqrt needs not be
        taken.
        Also, as K(x,x) is constant for each x it can be omitted. 

        Parameters
        ----------
        X: ndarray of shape(n_samples, n_clusters)
            data which should be assigned labels
        centers: ndarray of shape(n_clusters)
        
        Returns
        -------

        labels: ndarray of shape(n_samples)
            labels[i] == cluster X[i] is assigned to 
        '''
        X_center_kernel = self.kernel_wrapper(X, centers)
        dists_to_centers = np.zeros((len(X), self.n_clusters)) 
        for cluster in range(self.n_clusters):
            dists_to_centers[:, cluster] = (-2 * X_center_kernel[:, cluster]
                                         + self.kernel_wrapper(centers[cluster]))
        return np.asarray(np.argmin(dists_to_centers, axis=1), dtype=np.int_)

    def lloyd(self, kernel_matrix, labels):
        '''
        Computes a run of lloyd's algorithm

        Distances are actually sqared distances, as this is kmeans 
        objective function.
        As mentionend before, inner_sums and cluster_sizes 
        are computed anyways and are stored to safe time when 
        predicting.
        Utilizes cython-written update_lloyd function to do
        the heavy lifting.
        More details in update_lloyd

        Parameters
        ----------
        kernel_matrix: ndarray of shape(n_samples, n_samples)
            kernel matrix of the data to fit
        labels: ndarray of shape(n_samples)
            starting labels

        Returns:
            labels: ndarray of shape(n_samples)
                labels for each datapoint
            quality: float
                final quality score.
            inner_sums:
                mathematical term to ease prediction
                more info in class docstring
            cluster_sizes:
                size of each cluster
                returned to ease prediction
        '''
        # Init with -INF so that first iteration cannot converge because
        # quality does not change.
        quality = np.NINF
        for it in range(self.max_iter):
            labels, cluster_sizes = fill_empty_clusters(labels, self.n_clusters, rng=self.rng)
            sq_distances = self._build_starting_distance(kernel_matrix)
            sq_distances, inner_sums, cluster_sizes = update_lloyd(
                sq_distances, kernel_matrix, labels, 
                self.n_clusters, cluster_sizes)
            labels_old = labels
            labels = np.asarray(np.argmin(sq_distances, axis=1), dtype=np.int_)
            quality, converged = self._measure_iter(sq_distances, labels, labels_old, quality)
            self._out_verbose(it, quality, converged=converged)
            if converged:
                break
        
        return labels, quality, inner_sums, cluster_sizes
    


    def _out_verbose(self, iter, quality, converged):
        '''Checks if self.verbose and prints accordingly'''
        if not self.verbose:
            return
        if converged:
            print("Converged at iteration:", iter + 1, self.q_metric + ":", quality)
        elif self.tol == 0:
            print("Iteration:", iter + 1, self.q_metric + ":", "Not calculated (Tol==0)")
        else:
            print("Iteration:", iter + 1, self.q_metric + ":", quality)
    
       
    def _measure_iter(self, sq_distances, labels, labels_old, quality):
        '''
        Measures quality of iteration and checks for convergence

        If self.tol == 0 skips quality calculation except when labels did 
        not change to previous iteration, as then algorithm is finished.
        criteria for convergence are:
            abs(quality - quality_old) <=self.tol
            or all(labels == labels_old)
        
        Parameters
        ----------
        sq_distances: ndarray of shape(n_samples, n_clusters)
            squared distances from each sample to each cluster
        labels: ndarray of shape(n_samples)
            new labels for each sample
        labels_old: ndarray of shape(n_samples)
            labels from the previous iteration
        quality: float
            quality from the previous iteration
        
        Returns
        -------
        quality: float
            quality of current iteration
        converged: bool
            is the algorithm converged?
        '''
        converged = False
        if self.tol != 0:
            quality_old = quality
            quality = self.calc_q_metric(sq_distances, labels)
            if abs(quality - quality_old) <= self.tol:
                converged = True

        if all(labels_old == labels):
            if self.tol == 0:
                quality = self.calc_q_metric(sq_distances, labels)
            converged = True
        
        return quality, converged

    def calc_q_metric(self, sq_distances, labels):
        '''
        Calculates quality metric

        inertia is the sum of squared distances from each datapoint to
        its assigned center.
        silhouette is a heuristic that describes how significant the assignment
        of a point to its cluster is. Yields a value between -1 and 1.
        As elkan only guarantees to calculate the distances to the closest
        cluster center correctly, it is discouraged to measure it with 
        silhouette.

        Parameters
        ----------
        sq_distances: ndarray of shape(n_samples, n_features)
            square distances between points and centers
        labels: ndarray of shape(n_samples)
            labels for each datapoint

        Returns
        -------
        float, the calculated quality score
        '''
        if self.q_metric == "inertia":
            return sq_distances[range(sq_distances.shape[0]), labels].sum()
        elif self.q_metric == "silhouette":
            return self.calc_silhouette(sq_distances, labels)
        else:
            raise NotImplementedError(str(self.q_metric) + " quality metric not implemented")
            
    def calc_silhouette(self, sq_distances, labels):
        '''Calculates average silhouette of clustering
        
        sq_distances need to be exact for the two closest 
        centers for each datapoint (elkan does not provide this).
        Computes silhouette for every datapoint and averages it.
        
        Parameters
        ----------
        sq_distances: ndarray of shape(n_samples, n_clusters)
            squared distances from points to centers
        labels: ndarray of shape(n_samples)

        Returns
        -------
        Average silhouette of each datapoint
        '''
        return avg_silhouette(sq_distances, labels)
    
    def _build_starting_distance(self, kernel_matrix):
        '''
        Builds basis for computing square distances

        K(x, x),  K(x, x),  ....    K(x, x)
        K(x',x'), K(x',x'), ....    K(x',x')
        ...       ...       ....    ...
        ....
        '''
        # np.diag(kernel_matrix) returns [K(x,x), K(x', x'),...]
        # rest is for stacking correctly
        # ascontiguousarray is necessary as by transposing ndarray becomes f-ordered
        # c-order is necessary for cython.
        return np.ascontiguousarray(np.tile(np.diag(kernel_matrix), (self.n_clusters, 1)).T, dtype=np.double)

    def elkan(self, kernel_matrix, labels):
        '''
        Computes a slightly optimized Kernel-Kmeans heuristic

        This is _not_ the original elkan, but it is founded on
        the same observation regarding the triangle equality.
        Keeps record of how much centers have moved since the last
        exact computation of distance and only calculates distance 
        to a center if it is possible to have moved closer than the 
        currently assigned center.
        Relative speedup increases with n_clusters.

        Parameters
        ----------
        kernel_matrix: ndarray of shape(n_samples, n_samples)
        labels: ndarray of shape(n_samples)
            starting labels

        Returns:
            labels: ndarray of shape(n_samples)
            quality: float
            inner_sums:
                mathematical term to ease prediction
                more info in class docstring
            cluster_sizes:
                size of each cluster
                returned to ease prediction
        '''
        labels = np.array(labels, dtype=np.int_) 
        start_dists = self._build_starting_distance(kernel_matrix)
        center_dists = np.zeros((kernel_matrix.shape[0], self.n_clusters))
        quality = 0
        for it in range(self.max_iter):
            labels, sizes = fill_empty_clusters(labels, self.n_clusters, rng=self.rng)
            if it == 0:
                l_bounds, inner_sums, sizes = start_elkan(
                    start_dists, kernel_matrix, labels, 
                    self.n_clusters, sizes)
            else:
                l_bounds, inner_sums, sizes, center_dists = update_elkan(
                    kernel_matrix, l_bounds, center_dists, 
                    labels, labels_old, sizes, sizes_old,
                    inner_sums, self.n_clusters)

            labels_old = labels
            labels = np.array(np.argmin(l_bounds, axis=1), dtype=np.int_)
            quality, converged = self._measure_iter(l_bounds, labels, labels_old, quality)
            self._out_verbose(it, quality, converged=converged)
            if converged:
                break
            sizes_old = sizes
        return labels, quality, inner_sums, sizes

    def kmeanspp(self, X, kernel_matrix):
        '''
        Computes heuristic to get good initial centers

        If used externally, KKMeans instance has to be initialized
        properly (kernel, kwargs, n_clusters)
        Kmeans++:
            1. choose first center at random
            2. choose x as next center with probability of
                the normalized squared distance to its next 
                cluster center
            3. repeat 2 until k centers are found.
        Here, K(x,x) must not be omitted like in _assign_to_centers
        as the probabilities get normalized over the whole dataset.
        Returns labels and not centers as distances need to be computed 
        here anyhow.
        As the center are explicitely given, the distance calculations here
        are simpler.

        Parameters
        ----------
        X: ndarray of shape(n_samples, n_clusters)
            data which should be assigned labels.
            used to grab existing datapoints
        kernel_matrx: ndarray of shape(n_samples, n_samples)
            kernel matrix of X

        Returns
        -------
        ndarray of shape(n_samples)
            Labels for each datapoint
        '''
        #those will be square dists
        dists_to_centers = self._build_starting_distance(kernel_matrix)
        n_samples = X.shape[0]
        for cluster in range(self.n_clusters):
            if cluster == 0:
                # choose first center randomly
                index = self.rng.integers(low=0, high=n_samples)
            else:
                # careful, dists_to_centers not initialized with 0
                min_dist_each = np.amin(dists_to_centers[:, :cluster + 1], axis = 1) 
                probs = min_dist_each/min_dist_each.sum()
                index = self.rng.choice(n_samples, size=1, p=probs)
            center = X[index]
            inner_sum = self.kernel_wrapper(center)
            outer_sum = self.kernel_wrapper(X, center)
            # reshape necessary as kernel_wrapper has 2dim array output
            dists_to_centers[:, cluster] += (-2 * outer_sum + inner_sum).reshape(n_samples,)
            # eliminate flop errors
            dists_to_centers[np.isclose(dists_to_centers, 0)] = 0
        return np.asarray(np.argmin(dists_to_centers, axis=1), dtype=np.int_)


    def predict(self, X):
        '''
        Computes the closest cluster center for each x in X

        squared distance from datapoint x to center c:
            K(x,x) - 2K(x,c) + K(c,c)

        As sqrt is monotonic, 
            0 < x < y => 0 < sqrt(x) < sqrt(y)
        and therefore sqrt needs not be taken.
        As K(x,x) is constant for each x it does not
        influence the assignment and can consequently 
        be omitted.

        Parameters
        ----------
        X: ndarray
            the data which should be predicted
        
        Returns
        -------
        ndarray of shape(n_samples)
            labels for each datapoint in X
        '''
        if self.X_ is None:
            raise ValueError("KKMeans instance is not fitted!")
        kernel_matrix = self.kernel_wrapper(X, self.X_)
        sq_distances = calc_sq_distances(
            self.inner_sums_, self.cluster_sizes_,
            kernel_matrix, self.labels_, 
            self.n_clusters)
        return np.argmin(sq_distances, axis=1)



def visualize_kkm(data, labels):
    if len(data[0]) > 3:
        raise Exception("Dimensionality is too high for visualization")
    elif len(data[0]) == 1:
        plt.scatter(data, [0 for x in range(len(data))], c = labels)
    elif len(data[0]) == 2:
        plt.scatter(data[:,0], data[:,1], c = labels)
    elif len(data[0]) == 3:
        fig = plt.figure()
        ax = fig.add_subplot(projection = "3d")
        ax.scatter(data[:,0], data[:,1], data[:,2], c = labels)



import numpy as np
from cython.parallel import prange
from libc.math cimport sqrt 
def update_elkan(
        double[:, ::1] kernel_matrix, 
        double[:, ::1] l_bounds, 
        double[:, ::1] center_dists, 
        long[::1] labels, 
        long[::1] labels_old, 
        long[::1] sizes, 
        long[::1] sizes_old,
        double[::1] inner_sums,
        long n_clusters):
    '''
    updates the lower bounds estimation 

    computes necessary terms (on detailed info of inner_sums etc see 
    thesis) to update lower bounds estimation and then
    uses the algorithm described in KKMeans. (NOT vanilla elkan's
    algorithm!)

    parameters
    ----------
    kernel_matrix: ndarray of shape(n_samples, n_samples), dtype=np.double
    l_bounds: ndarray of shape(n_samples, n_clusters), dtype=np.double
        distance estimations
    center_dists: ndarray of shape(n_samples, n_clusters), dtype=np.double
        cumulative movements of each cluster relative for each datapoint
    labels: ndarray of shape(n_samples), dtype=np.int_
    labels_old: ndarray of shape(n_samples), dtype=np.int_
        labels of prev iteration, used to update center_dists
    sizes: ndarray of shape(n_samples), dtype=np.int_
    inner_sums: ndarray of shape(n_clusters)
    n_clusters: int

    returns
    -------
    l_bounds: ndarray of shape(n_samples, n_clusters)
        updated lower bounds
    inner_sums: ndarray of shape(n_clusters)
    sizes: ndarray of shape(n_clusters)
    center_dists: ndarray of shape(n_clusters, n_clusters)
        updated center_dists
    '''


    inner_sums_old = inner_sums
    inner_sums_mixed, inner_sums = _calc_inner_sums_mixed(kernel_matrix, labels, labels_old, 
                                                          n_clusters, return_inner_new=True)
    center_dists += _calc_center_dists(inner_sums, inner_sums_mixed, inner_sums_old, sizes, sizes_old)
    l_bounds = _est_lower_bounds(kernel_matrix, l_bounds, center_dists, labels, sizes, inner_sums)
    return np.asarray(l_bounds), np.asarray(inner_sums), np.asarray(sizes), np.asarray(center_dists)


def _est_lower_bounds(
        double[:, ::1] kernel_matrix, 
        double[:, ::1] l_bounds, 
        double[:, ::1] center_dists, 
        long[::1] labels, 
        long[::1] sizes, 
        double[::1] inner_sums):
    '''
    heuristic to update lower bounds

    at first, updates the lower bound from 
    each x to its assigned cluster-cetner c_x
    (and sets the relative movement from c_x to x to 0)
    then checks for each center c if it has moved enough
    to be possibly closer to x than c_x. if yes, the lower_bound
    from x to c is updated and the relative movement from c to x
    set to 0.
    '''
    cdef:
        Py_ssize_t i, j, labels_i
        double outer_sum

    assert l_bounds.shape[0] == labels.shape[0]
    assert l_bounds.shape[1] == inner_sums.shape[0]

    for i in prange(l_bounds.shape[0], nogil=True):
        labels_i = labels[i]
        outer_sum = _calc_outer_sum_single(kernel_matrix, i, labels_i, labels)
        # updating lower bounds from x to c_x
        l_bounds[i, labels_i] = (
            kernel_matrix[i, i] 
            - 2 * outer_sum / sizes[labels_i] 
            + inner_sums[labels_i] / sizes[labels_i]**2)
        center_dists[i, labels_i] = 0  # relative movement set to 0
        for j in range(l_bounds.shape[1]):
            if sqrt(l_bounds[i, j]) - center_dists[i, j] <= sqrt(l_bounds[i, labels_i]): 
                outer_sum = _calc_outer_sum_single(kernel_matrix, i, j, labels)
                l_bounds[i, j] = (kernel_matrix[i, i] 
                                 - 2 * outer_sum / sizes[j] 
                                 + inner_sums[j] / sizes[j]**2)
                center_dists[i, j] = 0
    return l_bounds


def _calc_center_dists(
        double[::1] inner_sums_new, 
        double[::1] inner_sums_mixed, 
        double[::1] inner_sums_old, 
        long[::1] sizes_new, 
        long[::1] sizes_old):
    '''
    mathematical helper function

    returns
    ------
    dists: ndarray of shape(n_clusters), dtype=np.double
        distances between new and old cluster centers
        dists[i] = ||center_i_new - center_i_old||
        NOT PAIRWISE!
    '''
    cdef:
        Py_ssize_t i
        Py_ssize_t n_clusters = inner_sums_new.shape[0]
        double[::1] dists = np.zeros(n_clusters)
        long new_size, old_size
        double mixed_sum, new_sum, old_sum

    assert n_clusters == inner_sums_mixed.shape[0] 
    assert n_clusters == inner_sums_old.shape[0]
    assert n_clusters == sizes_new.shape[0]
    assert n_clusters == sizes_old.shape[0]

    for i in range(n_clusters):
        new_size = sizes_new[i]
        old_size = sizes_old[i]
        mixed_sum = inner_sums_mixed[i]
        new_sum = inner_sums_new[i]
        old_sum = inner_sums_old[i]
        dists[i] = sqrt(new_sum / new_size**2 
                        - 2*mixed_sum / (new_size * old_size) 
                        + old_sum / old_size**2)
    
    return np.asarray(dists) 
    



def _calc_inner_sums_mixed(
        double[:, ::1] kernel_matrix, 
        long[::1] labels, 
        long[::1] labels_old, 
        long n_clusters, 
        return_inner_new=True):
    '''
    Mathematical helper function 

    Notes
    -----
    sums_mixed:
        sums_mixed[c, c'] = sum(K(x,y)) for x in c for y in c' 
        when return it is summed over the columns. then,
        sums_mixed[c] = sum(K(x,y) for x in c_i_new for y in c_i_old))
        (therefore sums_mixed, as it is pairwise over old and new centers)
    '''
    cdef:
        Py_ssize_t size = kernel_matrix.shape[0]
        double[:, ::1] sums_mixed = np.zeros((size, n_clusters))
        double[:, ::1] sums_new = np.zeros((size, n_clusters))
        Py_ssize_t i, j
        long labels_i
    
    assert size == kernel_matrix.shape[1]

    for i in prange(size, nogil=True):
        labels_i = labels[i]
        for j in range(size):
            if labels_i == labels[j]:
                sums_new[i, labels[j]] += kernel_matrix[i, j]
            if labels_i == labels_old[j]:
                sums_mixed[i, labels_old[j]] += kernel_matrix[i, j]
    if return_inner_new:
        return np.sum(sums_mixed, axis=0), np.sum(sums_new, axis=0)
    return np.sum(sums_mixed, axis=0)




def start_elkan(
        sq_distances, 
        double[:, ::1] kernel_matrix,
        long[::1] labels, 
        long n_clusters,
        long[::1] cluster_sizes):
    '''
    Init function for elkan

    functionally equivalent to update_lloyd, included here 
    to make code less dependent.
    Computes the square distances between all samples and all
    centers.

    Parameters
    ----------
    sq_distances: ndarray of shape(n_samples, n_clusters), dtype=np.double
        array on which the distances should be built upon. the one necessary
        here is provided by KKMeans.build_starting_distance()
    kernel_matrix: ndarray of shape(n_samples, n_samples), dtype=np.double
    labels: ndarray of shape(n_samples), dtype=np.int_
    n_clusters: int

    Returns
    -------
    sq_distances: ndarray of shape(n_samples, n_clusters), dtype=np.double
        exact, updated square distances
    inner_sums: ndarray of shape(n_clusters), dtype=np.double
        mathematical term used for prediction functionality
    cluster_sizes: ndarray of shape(n_clusters), dtype=np.int_
    '''
    cdef Py_ssize_t cluster

    
    sq_distances = np.asarray(sq_distances, dtype=np.double)
    outer_sums, inner_sums = _calc_sums_full(kernel_matrix, labels, n_clusters)
    
    for cluster in range(n_clusters):
        size = cluster_sizes[cluster]
        outer_sum = outer_sums[:, cluster]
        inner_sum = inner_sums[cluster] 
        sq_distances[:,  cluster] += (-2 * outer_sum / size 
                                     + inner_sum / size**2)
    return np.asarray(sq_distances), np.asarray(inner_sums), np.asarray(cluster_sizes)


def _calc_sums_full(double[:, ::1] kernel_matrix, long[::1] labels, long n_clusters):
    '''
    Mathematical helper function

    Returns
    -------
    outer_sums: ndarray of shape(n_samples, n_clusters)
        outer_sums[i, j] == sum(K(data[i], data[x])) for x in cluster j
        In words, the sum of the kernel of datapoint X[i] and all datapoints
        in cluster number j.
    inner_sums: ndarray of shape(n_clusters)
        "inner 'distance' sum"; 
        inner_sums[i] = sum(K(x,y)) for all x, y in cluster i
    '''
    cdef:
        Py_ssize_t rows = kernel_matrix.shape[0]
        Py_ssize_t cols = kernel_matrix.shape[1]
        double[:, ::1] inner_sum = np.zeros((rows, n_clusters), dtype=np.double)
        double[:, ::1] outer_sum = np.zeros((rows, n_clusters), dtype=np.double)
        Py_ssize_t i,j
        int label_i, label_j
    for i in prange(rows, nogil=True):
        label_i = labels[i]
        for j in range(cols):
            label_j = labels[j]
            outer_sum[i, label_j] += kernel_matrix[i, j]      
            if label_i == label_j:
                inner_sum[i, label_j] += kernel_matrix[i, j]
    return np.asarray(outer_sum), np.sum(inner_sum, axis=0)


cpdef double _calc_outer_sum_single(
        double[:, ::1] kernel_matrix,
        long elem_index, 
        long cluster_index, 
        long[::1] labels) nogil:
    '''
    mathematical helper function
    
    given sample x and cluster c:
    computes sum(K(x, y)) for all y in c 
    '''
    cdef:
        double outer_sum = 0.
        Py_ssize_t i
        Py_ssize_t size = len(labels)
    for i in range(size):
        if labels[i] == cluster_index:
            outer_sum += kernel_matrix[elem_index, i]
    return outer_sum



import numpy as np
from cython.parallel import prange

cdef extern from "float.h":
    cdef double DBL_MAX

def avg_silhouette(double[:, ::1] distances, long[::1] labels):
    '''returns the average silhouette of a dataset
    '''
    cdef:
        Py_ssize_t n_samples = distances.shape[0]
        Py_ssize_t n_clusters = distances.shape[1]
        double[::1] silhouettes = np.zeros(n_samples)
        Py_ssize_t i, j
        long labels_i
        double a, b, max_a_b

    if n_clusters <= 1:
        raise ValueError("Must have at least two clusters for silhouette")
    for i in prange(n_samples, nogil=True):
        labels_i = labels[i]
        a = distances[i, labels_i]
        b = DBL_MAX
        for j in range(n_clusters):
            if j == labels_i:
                continue
            b = min(distances[i, j], b)
        max_a_b = max(a,b)
        if max_a_b == 0:
            # if max_a_b = 0, a == b == 0 as dists always positive 
            silhouettes[i] = 0
            continue
        silhouettes[i] = (b - a) / max(a,b) 
    return np.sum(silhouettes)/len(silhouettes)

def calc_silhouettes(double[:, ::1] distances, long[::1] labels):
    '''
    calculates the silhouette of each datapoint

    For more details see KKMeans class docstring

    Parameters
    ----------
    distances: ndarray of shape(n_samples, n_clusters), dtype=np.double
        distances (or any measurement on a ratio scale) from each sample
        to each center
    labels: ndarray of shape(n_samples)
    
    Returns
    -------
    silhouettes: ndarray of shape(n_samples)
        the silhouette of each sample
    '''
    cdef:
        Py_ssize_t n_samples = distances.shape[0]
        Py_ssize_t n_clusters = distances.shape[1]
        double[::1] silhouettes = np.zeros(n_samples)
        Py_ssize_t i, j
        long labels_i
        double a, b, max_a_b

    if n_clusters <= 1:
        raise ValueError("Must have at least two clusters for silhouette")
    for i in prange(n_samples, nogil=True):
        labels_i = labels[i]
        a = distances[i, labels_i]
        b = DBL_MAX
        for j in range(n_clusters):
            if j == labels_i:
                continue
            b = min(distances[i, j], b)
        max_a_b = max(a,b)
        if max_a_b == 0:
            # if max_a_b = 0, a == b == 0 as dists always positive 
            silhouettes[i] = 0
            continue
        silhouettes[i] = (b - a) / max(a,b) 
    
    return np.asarray(silhouettes) 


cdef double max(double a, double b) nogil:
    if a > b:
        return a
    return b

cdef double min(double a, double b) nogil:
    if a < b:
        return a
    return b






import numpy as np
from cython.parallel import prange

def fill_empty_clusters(long[::1] labels, long n_clusters, return_sizes=True, rng=None):
    '''
    Utility to fill empty clusters

    Calculates cluster sizes via labels, if any size is zero
    a random element is assigned to empty cluster.
    EG cluster 0 has size 0, labels[random_index] = 0.
    At the beginning a new rng-generator is created to ensure
    predictable assignments. 
    As sizes must be calculated, can be used as alternative to
    calc_sizes

    Parameters
    ----------
    labels: ndarray of shape(n_samples), dtype=np.int_
    n_clusters: int
    return_sizes: bool
    rng: int
        seed used for random number generator

    Returns
    -------
    labels: ndarray of shape(n_samples), dtype=np.int_
        corrected labels so that no cluster is empty.
        if this has been fulfilled at the beginning,
        nothing changes.
    cluster_sizes: ndarray of shape(n_clusters), dtype=int_, optional
    '''

    cdef Py_ssize_t i, index
    cdef long n_samples = labels.size
    
    if n_samples < n_clusters:
        raise ValueError("n_samples must be > n_clusters")

    rng = np.random.default_rng(rng)

    while True:
        cluster_sizes = np.asarray(calc_sizes(labels, n_clusters), dtype=np.int_)
        # equal to: indices=[c for c in range(n_clusters) if cluster_sizes[c]==0]
        empty_cluster_indices = np.flatnonzero(cluster_sizes == 0)
        amount_empty_clusters = empty_cluster_indices.size
        if(amount_empty_clusters == 0):
            break
        for i in range(amount_empty_clusters):
            #assign every empty cluster one element
            print("Warning! Empty cluster encountered, consider using different n_cluster. Random element assigned to emtpy cluster")
            index = rng.integers(n_samples)
            labels[index] = empty_cluster_indices[i]
    if return_sizes:
        return np.asarray(labels, dtype=np.int_), np.asarray(cluster_sizes, dtype=np.int_)
    return np.asarray(labels, dtype=np.int_)


cpdef long[::1] calc_sizes(long[::1] labels, long n_clusters):
    '''calculates the size of every cluster'''
    cdef:
        Py_ssize_t size = labels.shape[0] 
        Py_ssize_t i
        long[::1] sizes = np.zeros(n_clusters, dtype=np.int_)
    for i in range(size):
        sizes[labels[i]] += 1
    return sizes


def calc_sq_distances(
        inner_sums,
        cluster_sizes,
        double[:, ::1] kernel_matrix,
        long[::1] labels,
        long n_clusters):
    '''
    computes sq_dists form every point to every cluster

    Intended to be used for predictions, as inner_sums, 
    cluster_sizes and labels are not extra computed here.

    Parameters
    ----------
    inner_sums: ndarray of shape(n_clusters), dtype=np.double
        inner_sums[i] = sum(K(x,y)) for all x, y in cluster i
    cluster_sizes: ndarray of shape(n_clusters), dtype=np.int_
    kernel_matrix: ndarray of shape(n_samples, n_pred_samples), dtype=np.double
        kernelmatrix between the fitted data and the samples to predict
    labels: ndarray of shape(n_samples), dtype=np.int_
    n_clusters: int

    Returns
    -------
    distances: ndarray of shape(n_pred_samples, n_clusters)
        sq distances of each sample to predict to each center
    '''
    outer_sum_full = np.asarray(_calc_outer_sums(kernel_matrix, labels, n_clusters))
    sq_distances = np.zeros((kernel_matrix.shape[0], n_clusters))
    for cluster in range(n_clusters):
        size = cluster_sizes[cluster]
        outer_sum = outer_sum_full[:, cluster]
        inner_sum = inner_sums[cluster]
        sq_distances[:, cluster] += (-2 * outer_sum / size 
                                    + inner_sum / size**2)
    return np.asarray(sq_distances)


cpdef double[:, ::1] _calc_outer_sums(
        double[:, ::1] kernel_matrix, 
        long[::1] labels, 
        long n_clusters):
    '''
    Mathematical helper function
    
    outer_sums[i, j] == sum(K(data[i], data[x])) for x in cluster j
    In words, the sum of the kernel of datapoint X[i] and all datapoints
    in cluster number j.
    '''
    cdef:
        Py_ssize_t rows = kernel_matrix.shape[0]
        Py_ssize_t cols = kernel_matrix.shape[1]
        double[:, ::1] outer_sums = np.zeros((rows, n_clusters), dtype=np.double)
        Py_ssize_t i,j
    for i in prange(rows, nogil=True):
        for j in range(cols):
            outer_sums[i, labels[j]] += kernel_matrix[i, j]    
    return outer_sums

cpdef double[::1] _calc_inner_sums(
        double[:, ::1] kernel_matrix, 
        long[::1] labels, 
        long n_clusters):
    '''
    just used for measurements
    '''
    cdef:
        Py_ssize_t rows = kernel_matrix.shape[0]
        Py_ssize_t cols = kernel_matrix.shape[1]
        double[:, ::1] inner_sums = np.zeros(shape=(rows, n_clusters), dtype=np.double)
        Py_ssize_t i,j
        long labels_i
    for i in prange(rows, nogil=True):
        labels_i = labels[i]
        for j in range(cols):
            if labels[j] == labels_i:
                inner_sums[i, labels_i] += kernel_matrix[i, j]    
    return np.sum(inner_sums, axis=0)



import numpy as np
from cython.parallel import prange
from libc.math cimport exp, tanh, fabs

def build_kernel_matrix(X, Y=None, kernel="linear", gamma=1, c_0=0, d=3, variance=1):
    '''
    wrapper function to build kernel matrix

    Parameters
    ----------
    X: ndarray of shape(n_samples, n_features), dtype=np.double
        data to build kernel matrix upon
    Y: ndarray of shape(*, n_features), dtype=np.double, optional
        If None, Y=X. data to compute kernel matrix 
        against, eg kernel_matrix[i,j] = K(X[i], Y[j])
        Y=X leads to expected behaviour
    kernel: {"linear", "rbf", "gaussian", "polynomial", "sigmoid", "laplacian"}
            default="linear"
    gamma: double, default=1
        param for kernel
    c_0: double, default=0
        param for kernel
    d: double, default=3
        param for kernel
    variance: double, default=1
        param for gaussian kernel.
        is passed to rbf with gamma=1/variance
    
    Returns
    -------
    kernel_matrix: ndarray, dtype=np.double 
        of shape(len(X), len(X)) if Y=None
        of shape(len(X), len(Y)) else
    '''
    if Y is None:
        Y = X
    X, Y = _sanitize_data(X, Y)
    if kernel == "linear":
        return np.asarray(linear_kernel(X, Y))
    elif kernel == "rbf":
        return np.asarray(rbf_kernel(X, Y, gamma))
    elif kernel == "gaussian":
        if np.isclose(variance, 0, atol=1e-8):
            raise ValueError("Variance too close to 0 for gaussian kernel!")
        return np.asarray(rbf_kernel(X, Y, 1/variance))
    elif kernel == "polynomial":
        return np.asarray(polynomial_kernel(X, Y, d, gamma, c_0))
    elif kernel == "sigmoid":
        return np.asarray(sigmoid_kernel(X, Y, gamma, c_0))
    elif kernel == "laplacian":
        return np.asarray(laplacian_kernel(X, Y, gamma))
    else:
        raise NotImplementedError(str(kernel) + " kernel not implemented")


def _sanitize_data(X, Y):
    X, Y = _cast_to_ndarray(X, Y)
    X, Y = _check_dimensionality(X, Y)
    return X, Y

def _check_dimensionality(X, Y):
    if len(X.shape) == 0 or len(Y.shape) == 0:
        raise ValueError("X and Y need to be 1-d or 2-d (arraylikes)")
    if 0 in X.shape or 0 in Y.shape:
        raise ValueError("Data needs to have at least one sample and one dimension")
    if len(X.shape) == 1:
        X = np.array([X])
    if len(Y.shape) == 1:
        Y = np.array([Y])
    if X.shape[1] != Y.shape[1]:
        raise ValueError("Dimension mismatch")
    if len(X.shape) > 2 or len(Y.shape) > 2:
        raise ValueError("X and Y need to be 1-d or 2-d")
    return X, Y

def _cast_to_ndarray(X, Y):
    X, Y = np.array(X, dtype=np.double), np.array(Y, dtype=np.double)
    return X, Y


cpdef linear_kernel(double[:, ::1] X, double[:, ::1] Y):
    '''K(x,y) == x.T * y'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double dot_prod
        double[:, ::1] matrix = np.zeros((x_size, y_size))

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            dot_prod = 0
            for k in range(dim):
                dot_prod = dot_prod +  X[i,k] * Y[j, k]
            matrix[i,j]= dot_prod
    return matrix


cpdef rbf_kernel(double[:, ::1] X, double[:, ::1] Y, double gamma):
    '''K(x,y) == exp(-gamma * ||x - y||^2)'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double sq_euclidian
        double[:, ::1] matrix = np.zeros((x_size, y_size))
    
    if np.isclose(gamma, 0, atol=1e-6):
        raise ValueError("Gamma too close to 0 for rbf kernel")

    assert dim == Y.shape[1]

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            sq_euclidian = 0
            for k in range(dim):
                sq_euclidian = sq_euclidian + (X[i,k] - Y[j,k]) ** 2
            matrix[i,j] = exp(-gamma * sq_euclidian)
    return matrix  

cpdef sigmoid_kernel(double[:, ::1] X, double[:, ::1] Y, double gamma, double c_0):
    '''k(x,y) == tanh(gamma * (x.T * y) + c_0)'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double dot_prod
        double[:, ::1] matrix = np.zeros((x_size, y_size))

    assert dim==Y.shape[1]

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            dot_prod = 0
            for k in range(dim):
                dot_prod = dot_prod + X[i,k] * Y[j,k]
            matrix[i,j] = tanh(gamma * dot_prod + c_0)
    return matrix

cpdef polynomial_kernel(double[:, ::1] X, double[:, ::1] Y, long d, double gamma, double c_0):
    '''K(x,y) == (gamma * (x.T * y) + c_0)**d'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double dot_prod
        double[:, ::1] matrix = np.zeros((x_size, y_size))

    assert dim == Y.shape[1]

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            dot_prod = 0
            for k in range(dim):
                dot_prod = dot_prod + X[i,k] * Y[j,k]
            matrix[i,j] = (gamma * dot_prod + c_0) ** d
    return matrix


cpdef laplacian_kernel(double[:, ::1] X, double[:, ::1] Y, double gamma):
    '''K(x,y) == exp(-gamma * ||x - y||_1)'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double one_norm
        double[:, ::1] matrix = np.zeros((x_size, y_size))

    assert dim == Y.shape[1]

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            one_norm = 0
            for k in range(dim):
                one_norm = one_norm + fabs(X[i,k] - Y[j,k])
            matrix[i,j] = exp(-gamma * one_norm)
    return matrix

cpdef chi_squared_kernel(double[:, ::1] X, double [:, ::1] Y, double gamma):
    '''K(x, y) == exp(-gamma * sum_i((x[i] - y[i])^2)/(x[i] + y[i]))'''
    cdef:
        Py_ssize_t x_size = X.shape[0]
        Py_ssize_t y_size = Y.shape[0]
        Py_ssize_t dim = X.shape[1]
        Py_ssize_t i,j,k
        double summand
        double[:, ::1] matrix = np.zeros((x_size, y_size))

    assert dim == Y.shape[1]

    for i in prange(x_size, nogil=True):
        for j in range(y_size):
            summand = 0
            for k in range(dim):
                summand = summand + (X[i, k] - Y[j, k])**2 / (X[i, k] + Y[j, k])
            matrix[i, j] = summand

    return matrix


from setuptools import setup, Extension
from Cython.Build import cythonize


# # default
compile_args = ["-DCYTHON_WITHOUT_ASSERTIONS"]

# # msvc
# compile_args = ["/O2", "-DCYTHON_WITHOUT_ASSERTIONS", "/openmp"]

# # gcc
# compile_args = ["-O3", "-ffast-math", "-DCYTHON_WITHOUT_ASSERTIONS", "-fopenmp"]

compiler_directives = {
    "language_level": 3, 
    "boundscheck": False, 
    "wraparound": False, 
    "cdivision": True,
}

extensions = [
    Extension("KKMeans.utils", ["src/KKMeans/utils.pyx"], extra_compile_args=compile_args),
    Extension("KKMeans.lloyd", ["src/KKMeans/lloyd.pyx"], extra_compile_args=compile_args),
    Extension("KKMeans.elkan", ["src/KKMeans/elkan.pyx"], extra_compile_args=compile_args),
    Extension("KKMeans.kernels", ["src/KKMeans/kernels.pyx"], extra_compile_args=compile_args),
    Extension("KKMeans.quality", ["src/KKMeans/quality.pyx"], extra_compile_args=compile_args),
    Extension("KKMeans", ["src/KKMeans/KKMeans.py"])
]

# This is used for building the distribution for pipy
# https://stackoverflow.com/questions/73766918/creating-python-package-which-uses-cython-valueerror
package_data = {
    "KKMeans" : ["utils.pyx", "lloyd.pyx", "elkan.pyx", "kernels.pyx", "quality.pyx", "KKMeans.py"]
}

# variables set here (except ext_modules) and not in Pyproject.toml as setuptool-specifics is still in beta
# https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html
setup(ext_modules=cythonize(extensions, compiler_directives=compiler_directives),package_dir={"KKMeans":"src/KKMeans"}, package_data=package_data, include_package_data=True)